#include <stdio.h>
#include <time.h>
#include <stdlib.h>

#define LONG long long int

/********************************************************************************/
/*  Zp utilities                                                                */
/********************************************************************************/

#define ADD64s fftadd64s
#define SUB64s fftsub64s
#define NEG64s fftneg64s
#define MUL64s fftmul64s
#define SUBMUL64s fftsubmul64s
#define MOD64s fftmod64s
#define MIN64s fftmin64s

#define INV64s modinv64s

inline LONG ADD64s(LONG a, LONG b, LONG p) { LONG t; t = (a-p)+b; t += (t>>63) & p; return t; }
inline LONG SUB64s(LONG a, LONG b, LONG p) { LONG t; t = a-b; t += (t>>63) & p; return t; }
inline LONG NEG64s(LONG a, LONG p) { LONG t; t = -a; t += (t>>63) & p; return t; }
inline LONG MUL64s(LONG a, LONG b, LONG p)
{
         LONG q, r;
         __asm__ __volatile__(           \
         "       mulq    %%rdx           \n\t" \
         "       divq    %4              \n\t" \
         : "=a"(q), "=d"(r) : "0"(a), "1"(b), "rm"(p));
         return r;
}
inline LONG SUBMUL64s(LONG a, LONG b, LONG c, LONG p) { return MUL64s(SUB64s(a,b,p),c,p); }
inline LONG MOD64(LONG a,LONG p) { return(a % p); }
inline LONG MIN64s(LONG a, LONG b) { if(a<b) return a; else return b; }

LONG modinv64s( LONG c, LONG p );
LONG powmod64s( LONG a, LONG n, LONG p );

/********************************************************************************/
/*  Polynomial utilities                                                        */
/********************************************************************************/

void vecprint64s( LONG *A, int n );
void polcopy64s( LONG *A, int d, LONG *B );

/********************************************************************************/
/*  Polynomial routines                                                         */
/********************************************************************************/

int polsqr64s( LONG * A, LONG * C, int d, LONG p );
int polmul64s( LONG * A, LONG * B, LONG * C, int da, int db, LONG p );
int poldiv64s( LONG * A, LONG * B, int da, int db, LONG p );

/********************************************************************************/
/*  FFT utilities and routines                                                  */
/********************************************************************************/

LONG getomega64s( LONG p, LONG n ) {   
   LONG a,m,omega;
   m = (p-1)/n;
   if( (p-1)-n*m ) return 0; // there is no such omega
   for( a=2; a<n; a++ )
      if( powmod64s(a,(p-1)/2,p) == p-1 ) {
          omega = powmod64s( a, m, p );
          return omega;
      }
   return 0; // should not happen
}

void FFT64s1( int n, LONG *a, LONG *W, LONG p )
{  int i,n2;
   LONG s,t; 
   if( n==1 ) return;
   if( n==2 ) { s = ADD64s(a[0],a[1],p);
                t = SUB64s(a[0],a[1],p);
                a[0] = s; a[1] = t; return; }
   //p = 469762049;
   //p = 1811939329;
   //p = 2013265921;
   if( n==4 ) {
      LONG t1,t2,t3,t4;
      t1 = ADD64s(a[0],a[2],p);
      t2 = SUB64s(a[0],a[2],p);
      t3 = ADD64s(a[1],a[3],p);
      t4 = SUBMUL64s(a[1],a[3],W[1],p);
      a[0] = ADD64s(t1,t3,p); // (a0+a2)+(a1+a3)
      a[1] = ADD64s(t2,t4,p); // (a0-a2)+w(a1-a3)
      a[2] = SUB64s(t1,t3,p); // (a0+a2)-(a1+a3)
      a[3] = SUB64s(t2,t4,p); // (a0-a2)-w(a1-a3)
      return;
   }
   n2 = n/2;
   for( i=0; i<n2; i++ ) {
      t = SUBMUL64s(a[i],a[n2+i],W[i],p);
      s = ADD64s(a[i],a[n2+i],p);
      a[i] = s; 
      a[n2+i] = t;
   }
   FFT64s1(n2,a   ,W+n2,p);
   FFT64s1(n2,a+n2,W+n2,p);
   return;
}

void FFT64s2( int n, LONG *a, LONG *W, LONG p ) 
{ int i,n2;
  LONG s,t; 
  if( n==1 ) return;
  if( n==2 ) { s = ADD64s(a[0],a[1],p);
               t = SUB64s(a[0],a[1],p);
               a[0] = s; a[1] = t; return; }
  //p = 469762049;
  //p = 1811939329;
  //p = 2013265921;
  if( n==4 ) {
      LONG t1,t2,t3,t4;
      t1 = ADD64s(a[0],a[2],p);
      t2 = SUB64s(a[0],a[2],p);
      t3 = ADD64s(a[1],a[3],p);
      t4 = SUBMUL64s(a[1],a[3],W[1],p);
      a[0] = ADD64s(t1,t3,p); // (a0+a2)+(a1+a3)
      a[1] = ADD64s(t2,t4,p); // (a0-a2)+w(a1-a3)
      a[2] = SUB64s(t1,t3,p); // (a0+a2)-(a1+a3)
      a[3] = SUB64s(t2,t4,p); // (a0-a2)-w(a1-a3)
      return;
  }
  n2 = n/2;
  FFT64s2(n2,a   ,W+n2,p);
  FFT64s2(n2,a+n2,W+n2,p);
  for( i=0; i<n2; i++ ) {
      t = MUL64s(a[n2+i],W[i],p);
      s = ADD64s(a[i],t,p);      
      t = SUB64s(a[i],t,p);      
      a[i] = s;               
      a[n2+i] = t;            
  }
  return;
}

void polmod64s( int d, int n, LONG *A, LONG *B, LONG p )
{       int i;
        for( i=0; i<=d; i++ ) A[i] = MOD64(B[i],p);
        for(    ; i<n;  i++ ) A[i] = 0;
        return;
}

void MakeW64( int n, LONG w, LONG *W, LONG p )
{  int i,j,m,n2;
   LONG *H,*I;
   W[0] = 1; n2 = n/2;
   for( i=1; i<n2; i++ ) W[i] = MUL64s(w,W[i-1],p);
   for( H=W,I=W+n2, m=n2/2; m>0; m=m/2 ) { for( j=0; j<m; j++ ) I[j] = H[2*j]; H=I; I=I+m; }
   return;
}

void MakeWinv64( int n, LONG *W, LONG p )
{  int i,j,n4,n2,m;
   LONG t,*H,*I;
   if( n==2 ) return;
   n2 = n/2; n4 = n/4;
   for( i=1; i<n4; i++ ) { t = W[i]; W[i] = p-W[n2-i]; W[n2-i] = p-t; }
   W[n4] = p-W[n4];
   for( H=W,I=W+n2, m=n4; m>0; m=m/2 ) { for( j=0; j<m; j++ ) I[j] = H[2*j]; H=I; I=I+m; }
   return;
}

int FFTmul64s( LONG * a, LONG * b, LONG * c, int da, int db, LONG *T, LONG p )
{
   int i,n,dc;
   LONG *A, *B, *W;
   dc = da+db;
   for( n=1; n<=dc; n*=2 );                       //printf("n := %d;\n",n);
   LONG w = getomega64s(p,n);                     //printf("w := %lld;\n",w);
   A = T; B = T+n; W = T+2*n;
   MakeW64(n,w,W,p);                                //printf("W := "); vecprint32s(W,n);
   polmod64s(da,n,A,a,p);                         //printf("A := "); vecprint32s(A,n);
   FFT64s1(n,A,W,p);                                 //printf("F := "); vecprint32s(A,n);
   if( a==b ) B = A;
   else {
       polmod64s(db,n,B,b,p);                     //printf("B := "); vecprint32s(B,n);
       FFT64s1(n,B,W,p);                             //printf("F := "); vecprint32s(B,n);
   }
   for( i=0; i<n; i++ ) A[i] = MUL64s(A[i],B[i],p);  //printf("AB := "); vecprint32s(A,n);
   MakeWinv64(n,W,p);                               //printf("W := "); vecprint32s(W,n);
   FFT64s2(n,A,W,p);                                 //printf("C := "); vecprint32s(A,n);
   LONG t = INV64s(n,p); 
   for( i=0; i<=dc; i++ ) c[i] = MUL64s(t,A[i],p);
   while( dc>=0 && c[dc]==0 ) dc--;
   return dc;
}

/*  this function is not used so I comment this out

LONG polsqrKaratsuba64s( LONG *a, LONG *c, LONG da, LONG *T, LONG p ) 
{
   LONG i, da1, da2, n, dc, d1, d2, d3;
   LONG*a1, *a2;
   if( da<100 ) return polsqr64s(a,c,da,p);
//printf("K(%lld): a := ",da); polprint64s(a,da);
   da1 = da/2;
   n = da1+1;
   da2 = da-n;
   a1 = a;
   while( da1>=0 && a[da1]==0 ) da1--;
   a2 = a+n;                                      // a = a1 + a2 x^n
   polcopy64s(a1,da1,c);                          // c = a1
   dc = polsub64s(c,a2,c,da1,da2,p);              // c -= a2
   d3 = polsqrKaratsuba64s(c,T,dc,T+2*n,p);       // T = (a1-a2)^2
//printf("K(%d): (a1-a2)^2 = ",da); polprint32s(T,d3);
   for( i=0; i<=2*da; i++ ) c[i] = 0;             // c = 0
   d1 = polsqrKaratsuba64s(a1,c,da1,T+2*n,p);     // c += a1^2
   d2 = polsqrKaratsuba64s(a2,c+2*n,da2,T+2*n,p); // c += a2^2*x^2n
//printf("K(%d): a1^2 = ",da); polprint32s(c,d1);
//printf("K(%d): a2^2 = ",da); polprint32s(c+2*n,d2);
   d3 = polsub64s(T,c,T,d3,d1,p);                 // T -= a1^2
   d3 = polsub64s(T,c+2*n,T,d3,d2,p);             // T -= a2^2
   polsub64s(c+n,T,c+n,d3,d3,p);                  // c -= T*x^n
//printf("K(%d): amid = ",da); polprint32s(T,d3);
//printf("K(%d): cfin = ",da); polprint32s(c,2*da);
   for( dc=2*da; dc>=0 && c[dc]==0; dc-- );
   return dc;
}
*/

void VECCOPY64s( LONG *A, LONG *B, int n ) { int i; for( i=0; i<n; i++ ) B[i] = A[i]; }
void VECFILL64s( LONG *A, int n, LONG x ) { int i; for( i=0; i<n; i++ ) A[i] = x; }
void VECSMUL64s( LONG *A, int n, LONG x, LONG p ) { int i; for( i=0; i<n; i++ ) A[i] = MUL64s(A[i],x,p); }
void VECPROD64s( LONG *A, LONG *B, int n, LONG p ) { int i; for( i=0; i<n; i++ ) A[i] = MUL64s(A[i],B[i],p); }

int polsqrFFT64s( LONG *a, int da, int n, LONG *W, LONG *Winv, LONG p )
{
    int i,dc;
    LONG ni;
    dc = 2*da;
    while( n/2>dc ) { n=n/2; W = W+n; Winv = Winv+n; }
    for( i=da+1; i<n; i++ ) a[i] = 0;
    FFT64s1(n,a,W,p);
    for( i=0; i<n; i++ ) a[i] = MUL64s(a[i],a[i],p);
    FFT64s2(n,a,Winv,p);
    ni = modinv64s(n,p);
    for( i=0; i<=2*da; i++ ) a[i] = MUL64s(ni,a[i],p);
    for( dc=2*da; dc>=0 && a[dc]==0; dc-- );
    return dc;
}

void polinv64s( LONG *f, int n, LONG *y, LONG *T, LONG p )
{
    // compute y = f^(-1) mod p as a power series to O(x^n)
    // y needs to be of size n
    // T needs to be of size ceil(3/2n)
    int m,i;
    if( f[0]==0 ) { printf("no inverse\n"); return; }
    if( n==1 ) { y[0] = INV64s(f[0],p); return; }
    m = (n+1)/2; 
    polinv64s( f, m, y, T, p );
    polmul64s( y, f, T, m-1, n-1, p );
    for( i=0; i<m; i++ ) T[i] = NEG64s(T[i+m],p);
    polmul64s( y, T, T, m-1, m-1, p );
    for( i=0; i<m && i+m<n; i++ ) y[i+m] = T[i];
    return;
}

#define INVCUTOFF 800
void FFTpolinvmodrec64s( LONG *f, int d, int n, LONG *y, LONG *W, LONG *Winv, LONG *T, LONG p )
{
    int i,m,n2;
    LONG *Y,*F,*M,ni;
    if( n<INVCUTOFF ) { polinv64s( f, MIN64s(n,d), y, T, p ); return; }
    Y = T;
    F = T+n;
    n2 = n/2;
    FFTpolinvmodrec64s( f, d, n2, y, W+n2, Winv+n2, T, p );
    VECCOPY64s(y,Y,n2); VECFILL64s(Y+n2,n2,0);
    FFT64s1(n,Y,W,p);
    m = MIN64s(d,n); VECCOPY64s(f,F,m); VECFILL64s(F+m,n-m,0);
    FFT64s1(n,F,W,p);
    VECPROD64s(F,Y,n,p); //for( i=0; i<n; i++ ) F[i] = MUL(F[i],Y[i],p);
    FFT64s2(n,F,Winv,p);
    ni = modinv64s(n,p);
    VECSMUL64s(F,n,ni,p);
    // M = "middle product" is in second half of F
    M = F+n2; VECCOPY64s(M,F,n2); VECFILL64s(M,n2,0);
    FFT64s1(n,F,W,p);
    VECPROD64s( F, Y, n, p ); // for( i=0; i<n; i++ ) F[i] = MUL(F[i],Y[i],p);
    FFT64s2(n,F,Winv,p);
    VECSMUL64s(F,n,ni,p);
    for( i=0; i<n; i++ ) y[n2+i] = NEG64s(F[i],p);      // yk = yk + x^n2 F
    return;
}

void polsqrmodsetup64s(LONG *f, int d,
                  int n,
                  LONG *W, LONG *Winv, LONG *Ffr, LONG *Ffrinv, LONG *T,
                  LONG p ) {
    int i,dq;
    LONG w;
    //clock_t st,et;
    // W,Winv,Ffr,Ffrinv all of length n and assumed to be consecutive in memory
    // Compute fr = recip(f), frinv = 1/fr mod x^(dq+1) and FFT1( fr ) and FFT1( fri )
    dq = 2*(d-1)-d; // = d-2
    for( i=0; i<=d; i++ ) Ffr[d-i] = f[i];  
    w = getomega64s( p, n );                MakeW64(n,w,W,p);     //printf("W="); vecprint64s(W,n);
    for( i=0; i<n; i++ ) Winv[i] = W[i];    MakeWinv64(n,Winv,p); //printf("Winv="); vecprint64s(Winv,n);
    //st = clock();
    //polinv64s( Ffr, dq+1, Ffrinv, T, p );
    FFTpolinvmodrec64s( Ffr, dq+1, n, Ffrinv, W, Winv, T, p );
    //polmul64s(Ffr,Ffrinv,T,dq,dq,p);      printf("Check="); polprint64s(T,dq);
    //et = clock();
    //printf("polmodinv: time=%d ms\n",(et-st)/1000);
    for( i=d+1; i<n; i++ ) Ffr[i] = 0;      FFT64s1( n, Ffr, W, p );
    for( i=dq+1; i<n; i++ ) Ffrinv[i] = 0;  FFT64s1( n, Ffrinv, W, p );
    return;
}

int polsqrmod64s( LONG *s, LONG *f, int ds, int df,
                  LONG *Ffr, LONG *Ffri, 
                  LONG *W, LONG *Winv, int n,
                  LONG *T, LONG p )
{
    // Compute s^2 mod f via multiplication by fri = fr^(-1) mod x^df using the FFT
    // Precomputed are W, Winv, Ffr = FFT(fr,W,n,p) and Ffri = FFT(fr^(-1),W,n,p)
    // In the comments below the subscript r in fr denotes the reciprocal polynomial
    int i, dr, dq; 
    LONG ninv, *S2 = T, *Qr = T+n;
    dq = df-2;
    ninv = INV64s(n,p);
    // df=4, ds=2  s = [s0,s1,s2,0] f = [f0,f1,f2,f3,f4] then sr = [0,s2,s1,s0]
    for( i=0; i<=ds; i++ ) Qr[i] = s[i];
    for( ; i<df; i++ ) Qr[i] = 0;                           // Qr = s
    for( ds=df-1, i=0; i<=ds; i++ ) S2[i] = Qr[ds-i];       // printf("sr := "); polprint32s(S2,ds);
    for( ; i<n; i++ ) S2[i] = 0;                            // S2 = sr
    FFT64s1( n, S2, W, p );                                    // S2 = F(sr)
    for( i=0; i<n; i++ ) S2[i] = MUL64s(S2[i],S2[i],p);        // S2 = F(sr^2)
    for( i=0; i<n; i++ ) Qr[i] = S2[i] ;                    // Qr = F(sr^2)
    FFT64s2( n, Qr, Winv, p );                                 
    for( i=0; i<=dq; i++ ) Qr[i] = MUL64s(ninv,Qr[i],p);       // Qr = sr^2
    // printf("sr^2 = "); vecprint32s(Qr,n);
    for( ; i<n; i++ ) Qr[i] = 0;                            // Qr = s^2 mod x^(dq+1)
    //printf("sr^2 mod x^%d = ",df); vecprint32s(Qr,n);
    FFT64s1( n, Qr, W, p );                                    // Qr = F(s^2 mod x^(dq+1))
    for( i=0; i<n; i++ ) Qr[i] = MUL64s(Qr[i],Ffri[i],p);      // Qr = F( that / fr )
    FFT64s2( n, Qr, Winv, p );                                 // Qr = qr
    for( i=0; i<=dq; i++ ) Qr[i] = MUL64s(ninv,Qr[i],p);
    for( ; i<n; i++ ) Qr[i] = 0;                            // Qr = qr mod x^(dq+1)
    //printf("qr = "); polprint32s(Qr,dq);
    FFT64s1( n, Qr, W, p );                                    // Qr = F(qr)
    for( i=0; i<n; i++ )                                    // Qr = F( sr^2 - qr fr )
        Qr[i] = SUB64s(S2[i],MUL64s(Qr[i],Ffr[i],p),p);
    FFT64s2( n, Qr, Winv, p );                                 // Qr = n (sr^2 - qr fr)
    for( i=0; i<df; i++ ) T[df-i-1] = Qr[i+df-1];           // Qr = n (s^2 - q f)
    for( dr=df-1; dr>=0 && T[dr]==0; dr-- );
    for( i=dr; i>=0; i-- ) T[i] = MUL64s(ninv,T[i],p);         // T = s^2 - q f
    //printf("r = "); polprint32s(T,df-1);
    return dr;
}

/* C(x) := A(x)^n mod B(x) mod p;  0<=deg(A)<deg(B) and T must be of size 6N  */
int FFTpolpowmod64s( LONG * A, LONG n, LONG * B, int da, int db, LONG *C, LONG *T, LONG p )
{
    int b[63],dc,k,K,L,M;
    if( n==0 ) { C[0] = 1; return 0; }
    for( k=0; n>0; k++ ) { b[k]=n&1; n=n/2; }
    if( da>=db ) da = poldiv64s(A,B,da,db,p);                   // reduce A mod B first
    polcopy64s(A,da,C);
    dc = da;

    int N;
    LONG *W, *Winv, *Ffr, *Ffrinv, *R;
    for( N=1; N<2*db-1; N *= 2 );
    //printf(" FFTpolpowmod: N=%lld  alpha=%lld  p=%lld\n",N,alpha,p);
    W = T; Winv = T+N; Ffr = T+2*N; Ffrinv = T+3*N; R = T+4*N;
    polsqrmodsetup64s( B, db, N, W, Winv, Ffr, Ffrinv, R, p );

    k --; K = 0; L = 0; M = 0;
    while( k>0 ) { k--;
       // Main step: compute C := C^2 mod B in Zp[x]
       if( dc<400 && 2*dc<db ) { dc = polsqr64s(C,R,dc,p); polcopy64s(R,dc,C); }
       else if( 2*dc<db ) { L++; dc = polsqrFFT64s(C,dc,N,W,Winv,p); }
       else { M++; dc = polsqrmod64s(C,B,dc,db,Ffr,Ffrinv,W,Winv,N,R,p);
              polcopy64s(R,dc,C);  }
       K++;
       //printf("deg(C) = %d; C = ",dc); //polprint32s(C,dc);
       if( b[k]==1 ) {                                          //printf(" b[%d]=%d \n", k, b[k] );
           dc = polmul64s(A,C,R,da,dc,p);                       //printf("deg(R) = %d; R = ",dc); polprint32s(R,dc);
           dc = poldiv64s(R,B,dc,db,p);
           polcopy64s(R,dc,C);                                  //printf("deg(C) = %d; C = ",dc); polprint32s(C,dc);
       }
    }
if( db > 10000 ) {
    printf("We did %d squarings mod x^%d+... in total\n",K,db);
    printf("We used classical x for %d squarings \n",K-L-M);
    printf("We used the FFT for %d simple squarings \n",L);
    printf("We used the FFT for %d mod squarings of them\n",M);
}
    return dc;
}
