/* unsigned 128-bit integer math */
/* Roman Pearce  CECM/SFU   2015 */
/* contact: rpearcea@cecm.sfu.ca */

/* code reciprocal division/mulmod */
/* speed up generic C code add/mul */
/* recode 256/128 division routine */
/* remove dependency on __int128_t */

#define INT128_ASSEMBLY 1	/* use x86-64 assembly code */
#define INT128_EXTENDED 1	/* functions for __int128_t */

#define INT32 int
#define INT64 long long
#define UINT32 unsigned int
#define UINT64 unsigned long long

void add128(UINT64 *x, UINT64 *y);		/* x += y */
void add256(UINT64 *x, UINT64 *y);		/* x += y */
void sub128(UINT64 *x, UINT64 *y);		/* x -= y */
void sub256(UINT64 *x, UINT64 *y);		/* x -= y */

void mul128(UINT64 *x, UINT64  a, UINT64  b);	/* x = a*b where a and b are 64-bits */
void mul256(UINT64 *x, UINT64 *y, UINT64 *z);	/* x = y*z where y and z are 128-bit */

void div128(UINT64 *x, UINT64  a);		/* x = [ x / a : x % a ] */
void div256(UINT64 *x, UINT64 *y);		/* x = [ x / y : x % y ] */

UINT64 mulmod64(UINT64  a, UINT64  b, UINT64  p);
void  mulmod128(UINT64 *x, UINT64 *y, UINT64 *z, UINT64 *p);

/* division routines that use reciprocal multiplication */
/* from Moller and Granlund: Improved Division by Invariant Integers */

typedef struct {
	UINT64 s;	/* shift */
	UINT64 v;	/* reciprocal of d */
	UINT64 d0;	/* divisor shifted up */
	UINT64 d1;
} recint;

recint recip1(UINT64  p);			/* reciprocal of  64-bit p */
recint recip2(UINT64 *p);			/* reciprocal of 128-bit p */

void   rec128(UINT64 *x, recint v);		/* reciprocal version of div128 */
void   rec256(UINT64 *x, recint v);		/* reciprocal version of div256 */

UINT64 mulrec64(UINT64  a, UINT64  b, recint  v);
void  mulrec128(UINT64 *x, UINT64 *y, UINT64 *z, recint v);
void fma128(UINT64 *x, UINT64  a, UINT64  b);


